/*-------------------------------------------------------------------------------
 *
 *   Copyright (C) 2006,2007,2008,2009,2010 AVM GmbH <fritzbox_info@avm.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation version 2 of the License.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *-------------------------------------------------------------------------------
 *
 *-------------------------------------------------------------------------------
 * kspeedtest: AVM Kernel Speedtest
 *-------------------------------------------------------------------------------
 *
 * After module load kspeedtest can be controlled via proc interface
 * /proc/net/kspeedtest
 *
 * echo <command> > /proc/net/kspeedtest/control
 *
 * Commands:
 *
 * - start tcp [port] 	- Start TCP server (default Port 4711)
 * - start udp [port] 	- Start UDP server (default Port 2325)
 * - stop tcp			- Stop TCP server
 * - stop udp			- Stop UDP server
 * - stop all			- Stop all servers
 * - verbose			- Print stats as kernel message
 * - noverbose  		- Turn off verbose
 *
 * 
 * Show stats (Received Bytes, BPS, AVG BPS):
 *
 * cat /proc/net/kspeedtest/stats
 *
 * Show connections
 *
 * cat /proc/net/kspeedtest/connections
 *
 * vim: expandtab fileencoding=utf-8
 *
 */

#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kthread.h>

#include <linux/errno.h>
#include <linux/types.h>

#include <linux/netdevice.h>
#include <linux/ip.h>
#include <linux/in.h>
#include <linux/in6.h>
#include <linux/proc_fs.h>

#include <linux/delay.h>
#include <linux/socket.h>
#include <linux/tcp.h>
#include <net/sock.h>
#include <net/tcp_states.h>

#define DEFAULT_UDP_PORT 2325
#define DEFAULT_TCP_PORT 4711
#define MODULE_NAME "kspeedtest"
//#define INADDR_SEND ((unsigned long int)0x7f000001) /* 127.0.0.1 */
#define INADDR_SEND INADDR_LOOPBACK

#undef NO_AUTO_RWIN

const struct in6_addr in6addr_any = IN6ADDR_ANY_INIT;

MODULE_LICENSE("GPL");

static struct timer_list ktimer;

static void Second(unsigned long param);

static struct kspeedtest {
   int secondtimer;
   int verbose;
   unsigned udp_port;
   unsigned tcp_port;
   unsigned int seconds;
   unsigned int packets;
   unsigned long total_packets;
   unsigned long bytes;
   unsigned long long total_bytes;
   unsigned int current_bps;
   unsigned int current_pps;
   unsigned int avg_bps;
   unsigned int avg_pps;
} kspeedtest;

enum client_state {
   client_init,
   client_listen,
   client_connected
};

struct kthread_t
{
    struct task_struct *thread;
    struct socket *sock;
    struct socket *client;
	enum client_state client_state;
    struct sockaddr_in6 addr;
    int running;
};

struct kthread_t *udp_thread = NULL;
struct kthread_t *tcp_thread = NULL;

/* function prototypes */
static int udp_receive(struct socket *sock, struct sockaddr_in6 *addr, unsigned char *buf, int len);
#if 0
static int udp_send(struct socket *sock, struct sockaddr_in6 *addr, unsigned char *buf, int len);
#endif
static void __init kspeedtest_proc_init(void);
static void __init kspeedtest_proc_exit(void);

typedef int my_fprintf(void *, const char *, ...)
#ifdef __GNUC__
	        __attribute__ ((__format__(__printf__, 2, 3)))
#endif
	        ;


static void Second(unsigned long param)
{
	u64 x64;
	kspeedtest.seconds++;
	kspeedtest.total_bytes += kspeedtest.bytes;
	kspeedtest.total_packets += kspeedtest.packets;
	kspeedtest.current_bps = kspeedtest.bytes*8;
	x64 = kspeedtest.total_bytes*8;
	do_div(x64, kspeedtest.seconds);
	kspeedtest.avg_bps = x64;
	kspeedtest.current_pps = kspeedtest.packets;
	kspeedtest.avg_pps = kspeedtest.total_packets / kspeedtest.seconds;
    ktimer.expires = jiffies + HZ;
	add_timer(&ktimer);

	if (kspeedtest.verbose && kspeedtest.bytes)
		printk(KERN_ERR MODULE_NAME": Bytes %llu \t BPS %u AVG %u Packets %lu PPS %u AVG %u\n", 
			kspeedtest.total_bytes, kspeedtest.current_bps, 
			kspeedtest.avg_bps,	kspeedtest.total_packets, 
			kspeedtest.current_pps, kspeedtest.avg_pps);

	kspeedtest.bytes = 0;
	kspeedtest.packets = 0;
}

static void kspeedtest_show_stats(my_fprintf fprintffunc, void *arg)
{
	(*fprintffunc)(arg, "Bytes %llu \t BPS %u \t AVG %u Packets %lu PPS %u AVG %u\n",
			kspeedtest.total_bytes, kspeedtest.current_bps, 
			kspeedtest.avg_bps,	kspeedtest.total_packets, 
			kspeedtest.current_pps, kspeedtest.avg_pps);
}

/***************************************************************************/
/*                             UDP                                         */
/***************************************************************************/

static int kspeedtest_udp_thread(void)
{
	int size, err;
	int bufsize = 1500;
	unsigned char buf[bufsize+1];

	udp_thread->running = 1;
	current->flags |= PF_NOFREEZE;

	daemonize(MODULE_NAME);
	allow_signal(SIGKILL);

	/* create a udp socket */
	if ((err = sock_create_kern(AF_INET6, SOCK_DGRAM, IPPROTO_UDP, &udp_thread->sock)) < 0) {
		printk(KERN_ERR MODULE_NAME": Could not create a datagram socket, error = %d\n", -ENXIO);
		goto out;
	}

	memset(&udp_thread->addr, 0, sizeof(struct sockaddr));
	udp_thread->addr.sin6_family      = AF_INET6;
	udp_thread->addr.sin6_addr 		  = in6addr_any;
	udp_thread->addr.sin6_port        = 
		htons(kspeedtest.udp_port ? kspeedtest.udp_port : DEFAULT_UDP_PORT);

	if ( (err = udp_thread->sock->ops->bind(udp_thread->sock, 
					(struct sockaddr *)&udp_thread->addr, sizeof(struct sockaddr_in6) ) ) < 0) {
		printk(KERN_ERR MODULE_NAME": Could not bind or connect to udp socket, error = %d\n", -err);
		goto close_and_out;
	}

	printk(KERN_INFO MODULE_NAME": listening on port %d\n", 
			ntohs(udp_thread->addr.sin6_port));

	for (;;) {
		
		if (signal_pending(current))
			break;

		memset(&buf, 0, bufsize+1);
		size = udp_receive(udp_thread->sock, &udp_thread->addr, buf, bufsize);

		if (size < 0) {
			printk(KERN_ERR MODULE_NAME": error getting datagram, sock_recvmsg error = %d\n", size);
		} else {
			kspeedtest.bytes += size;
			kspeedtest.packets++;
		}
	}

	close_and_out:
	sock_release(udp_thread->sock);
	udp_thread->sock = NULL;
	out:
	udp_thread->thread = NULL;
	udp_thread->running = 0;
	return 0;
}

#if 0
static int udp_send(struct socket *sock, struct sockaddr_in *addr, unsigned char *buf, int len)
{
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
	int size = 0;

	if (sock->sk==NULL)
	   return 0;

	iov.iov_base = buf;
	iov.iov_len = len;

	msg.msg_flags = 0;
	msg.msg_name = addr;
	msg.msg_namelen  = sizeof(struct sockaddr_in);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	size = sock_sendmsg(sock,&msg,len);
	set_fs(oldfs);

	return size;
}
#endif

static int udp_receive(struct socket* sock, struct sockaddr_in6* addr, unsigned char* buf, int len)
{
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
	int size = 0;

	if (sock->sk==NULL) return 0;

	iov.iov_base = buf;
	iov.iov_len = len;

	msg.msg_flags = 0;
	msg.msg_name = addr;
	msg.msg_namelen  = sizeof(struct sockaddr_in);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	size = sock_recvmsg(sock,&msg,len,msg.msg_flags);
	set_fs(oldfs);

	return size;
}

static int start_udp_thread(void)
{
	if (udp_thread != NULL && udp_thread->thread != NULL) {
		printk(KERN_ERR MODULE_NAME": udp thread already started!\n");
		return -1;
	}
	if (udp_thread) kfree(udp_thread);
	udp_thread = kmalloc(sizeof(struct kthread_t), GFP_KERNEL);
	memset(udp_thread, 0, sizeof(struct kthread_t));
	udp_thread->thread = kthread_run((void *)kspeedtest_udp_thread, 
			                         NULL, MODULE_NAME);
	if (IS_ERR(udp_thread->thread)) 
	{
		printk(KERN_ERR MODULE_NAME": unable to start kernel thread\n");
		kfree(udp_thread);
		udp_thread = NULL;
		return -ENOMEM;
	}
	printk(KERN_INFO MODULE_NAME": kspeedtest_init: udp thread started \n"); 
	return 0;
}

static int stop_udp_thread(void)
{
	if (udp_thread == NULL) return 0;

	if (udp_thread->thread) {
	   (void)send_sig(SIGKILL, udp_thread->thread, 1);
	}

	if (udp_thread->thread==NULL)
		printk(KERN_ERR MODULE_NAME": no kernel thread to kill\n");
	else 
	{
		while(udp_thread->running) msleep(10);
		printk(KERN_INFO MODULE_NAME": successfully killed kernel udp thread\n");
	}
	kfree(udp_thread);
	udp_thread = NULL;
	return 0;
}

/***************************************************************************/
/*                             TCP                                         */
/***************************************************************************/

static int tcp_receive(struct socket* sock, struct sockaddr_in6* addr, unsigned char* buf, int len)
{
	struct msghdr msg;
	struct iovec iov;
	mm_segment_t oldfs;
	int size = 0;

	if (sock->sk==NULL) return 0;

	iov.iov_base = buf;
	iov.iov_len = len;

	msg.msg_flags = MSG_MORE;
	msg.msg_name = addr;
	msg.msg_namelen  = sizeof(struct sockaddr_in);
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;

	oldfs = get_fs();
	set_fs(KERNEL_DS);
	size = sock_recvmsg(sock,&msg,len,msg.msg_flags);
	set_fs(oldfs);

	return size;
}

static int tcp_handle_data(struct kthread_t *kthread)
{
   int err, size;
   int bufsize = 1500;
   unsigned char buf[bufsize+1];

   switch(kthread->client_state)
   {
   case client_init:
	   break;
   case client_listen:
	   {
#ifdef NO_AUTO_RWIN
		   int rwin = 256000;
#endif
		   err = sock_create_lite(AF_INET6, SOCK_STREAM, IPPROTO_TCP, &kthread->client);
		   if (err || !kthread->client) {
			   printk(KERN_ERR MODULE_NAME": error creating client socket");
			   return -1;
		   }

		   kthread->client->type = kthread->sock->type;
		   kthread->client->ops  = kthread->sock->ops;


		   err = kthread->sock->ops->accept(kthread->sock, kthread->client, 0);
		   if (err < 0) {
			   sock_release(kthread->client);
			   kthread->client = NULL;
			   printk(KERN_ERR MODULE_NAME": error accepting client connection");
			   return -2;
		   }
		   printk(KERN_INFO MODULE_NAME": client connection accepted \n ");
#ifdef NO_AUTO_RWIN
		   err = sock_setsockopt(kthread->client, SOL_SOCKET, SO_RCVBUF, 
			   (char *)&rwin, sizeof(rwin)); 
		   printk(KERN_INFO MODULE_NAME": setting rwin %d ret=%d\n", rwin, err);
#endif
		   kthread->client_state = client_connected;
		   kspeedtest.seconds = 0;
		   kspeedtest.total_bytes = 0;
		   kspeedtest.total_packets = 0;
	   }
	   break;
   case client_connected:
	   if (kthread->client->sk->sk_state == TCP_CLOSE_WAIT) {
		   sock_release(kthread->client);
		   kthread->client = NULL;
		   kthread->client_state = client_listen;
		   printk(KERN_INFO MODULE_NAME": client connection closed\n");
	   } else {
		   size = tcp_receive(kthread->client, &kthread->addr, buf, bufsize);

		   if (size < 0) {
			   printk(KERN_ERR MODULE_NAME": error getting bytes, sock_recvmsg error = %d\n", size);
		       sock_release(kthread->client);
		       kthread->client = NULL;
		       kthread->client_state = client_listen;
		       printk(KERN_INFO MODULE_NAME": client connection closed\n");
		   } else {
			   kspeedtest.bytes += size;
			   kspeedtest.packets++;
		   }
	   }
	   break;
   }
   return 0;
}


static int kspeedtest_tcp_thread(void)
{
	int err;
#ifdef NO_AUTO_RWIN
	int rwin = 256000;
#endif

	/* kernel thread initialization */
	tcp_thread->running = 1;
	current->flags |= PF_NOFREEZE;

	/* daemonize (take care with signals, after daemonize() they are disabled) */
	daemonize(MODULE_NAME);
	allow_signal(SIGKILL);

	/* create a tcp socket */
	if ((err = sock_create_kern(AF_INET6, SOCK_STREAM, IPPROTO_TCP, &tcp_thread->sock)) < 0) {
		printk(KERN_ERR MODULE_NAME": Could not create a stream socket, error = %d\n", -ENXIO);
		goto out;
	}

	memset(&tcp_thread->addr, 0, sizeof(struct sockaddr));
	tcp_thread->addr.sin6_family      = AF_INET6;
	tcp_thread->addr.sin6_addr        = in6addr_any;
	tcp_thread->addr.sin6_port        = 
		htons(kspeedtest.tcp_port ? kspeedtest.tcp_port : DEFAULT_TCP_PORT);

#ifdef NO_AUTO_RWIN
	err = sock_setsockopt(tcp_thread->sock, SOL_SOCKET, SO_RCVBUF, 
			(char *)&rwin, sizeof(rwin)); 
	printk(KERN_INFO MODULE_NAME":  rwin %d ret=%d\n", rwin, err);
#endif

	if ( (err = tcp_thread->sock->ops->bind(tcp_thread->sock, 
					(struct sockaddr *)&tcp_thread->addr, sizeof(struct sockaddr_in6) ) ) < 0) {
		printk(KERN_ERR MODULE_NAME": Could not bind or connect to tcp socket, error = %d\n", -err);
		goto close_and_out;
	}
	//tcp_thread->sock->sk->sk_reuse = 1;
	tcp_thread->sock->ops->listen(tcp_thread->sock, 48); //why 48?
	tcp_thread->client_state = client_listen;

	printk(KERN_INFO MODULE_NAME": listening on port %d\n", 
			ntohs(tcp_thread->addr.sin6_port));

	while (err >= 0) {
		if (signal_pending(current)) 
			break;

		err = tcp_handle_data(tcp_thread);
	}

close_and_out:
	sock_release(tcp_thread->sock);
	tcp_thread->sock = NULL;
out:
	if (tcp_thread->client) {
		sock_release(tcp_thread->client);
		tcp_thread->client = NULL;
	}
	tcp_thread->thread = NULL;
	tcp_thread->running = 0;
	return 0;
}

static int start_tcp_thread(void) 
{
	if (tcp_thread != NULL && tcp_thread->thread != NULL) {
		printk(KERN_ERR MODULE_NAME": tcp thread already started!\n");
		return -1;
	}
	if (tcp_thread) kfree(tcp_thread);
	tcp_thread = kmalloc(sizeof(struct kthread_t), GFP_KERNEL);
	memset(tcp_thread, 0, sizeof(struct kthread_t));
	tcp_thread->thread = kthread_run((void *)kspeedtest_tcp_thread, 
			                         NULL, MODULE_NAME);
	tcp_thread->client_state = client_init;
	if (IS_ERR(tcp_thread->thread)) 
	{
		printk(KERN_ERR MODULE_NAME": unable to start kernel thread\n");
		kfree(tcp_thread);
		tcp_thread = NULL;
		return -ENOMEM;
	}
	printk(KERN_INFO MODULE_NAME": kspeedtest_init: tcp thread started \n"); 
	return 0;
}

static int stop_tcp_thread(void)
{
	if (tcp_thread == NULL) return 0;

	if (tcp_thread->thread) {
	   (void)send_sig(SIGKILL, tcp_thread->thread, 1);
	}

	if (tcp_thread->thread==NULL)
		printk(KERN_ERR MODULE_NAME": no kernel thread to kill\n");
	else 
	{
		while(tcp_thread->running) msleep(10);
		printk(KERN_INFO MODULE_NAME": successfully killed kernel tcp thread\n");
	}
	kfree(tcp_thread);
	tcp_thread = NULL;
	return 0;
}

/*********************************************************************************/

static int __init kspeedtest_init(void)
{


	init_timer(&ktimer);
	ktimer.function = Second;
	ktimer.data = 0;
	ktimer.expires = jiffies + HZ;
	add_timer(&ktimer);
	printk(KERN_INFO MODULE_NAME": kspeedtest_init: second timer started \n"); 

#ifdef CONFIG_PROC_FS
	kspeedtest_proc_init();
#endif

	printk(KERN_INFO MODULE_NAME": kspeedtest_init done\n"); 

	return 0;
}

static void __exit kspeedtest_exit(void)
{
	del_timer_sync(&ktimer);

	stop_udp_thread();
	stop_tcp_thread();

#ifdef CONFIG_PROC_FS
	kspeedtest_proc_exit();
#endif

	printk(KERN_INFO MODULE_NAME": module unloaded\n");
}

#ifdef CONFIG_PROC_FS

/* -------------------------------------------------------------------- */

static int stats_show(struct seq_file *m, void *v)
{
	kspeedtest_show_stats((my_fprintf *)seq_printf, m);
	return 0;
}

static int stats_show_open(struct inode *inode, struct file *file)
{
	    return single_open(file, stats_show, PDE(inode)->data);
}

static const struct file_operations stats_show_fops = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
	.owner   = THIS_MODULE,
#endif
	.open    = stats_show_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release,
};

static int connections_show(struct seq_file *m, void *v)
{
	if (tcp_thread && tcp_thread->running) {
		seq_printf(m, "TCP %d %sconnected\n",
				kspeedtest.tcp_port ? kspeedtest.tcp_port : DEFAULT_TCP_PORT,
				tcp_thread->client_state == client_connected ? "" : "not ");
	} else {
		seq_printf(m, "TCP off\n");
	}
	if (udp_thread && udp_thread->running) {
		seq_printf(m, "UDP %d\n",
				kspeedtest.udp_port ? kspeedtest.udp_port : DEFAULT_UDP_PORT);
	} else {
		seq_printf(m, "UDP off\n");
	}
	return 0;
}

static int connections_show_open(struct inode *inode, struct file *file)
{
	    return single_open(file, connections_show, PDE(inode)->data);
}

static const struct file_operations connections_show_fops = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
	.owner   = THIS_MODULE,
#endif
	.open    = connections_show_open,
	.read    = seq_read,
	.llseek  = seq_lseek,
	.release = seq_release,
};


static int kspeedtest_write_cmds (struct file *file, const char *buffer,
                                  unsigned long count, void *data)
{
	char cmd[100];
	char *argv[10];
	int  argc = 0;
	char *ptr_cmd;
	char *delimitters = " \n\t";
	char *ptr_next_tok;

	if (count > 100) count = 100;

	memset((void *)&cmd[0], 0, sizeof(cmd));
	memset((void *)&argv[0], 0, sizeof(argv));

	if (copy_from_user(&cmd, buffer, count))
		return -EFAULT;

	ptr_next_tok = &cmd[0];
	ptr_cmd = strsep(&ptr_next_tok, delimitters);
	if (ptr_cmd == NULL) return -1;

	do 
	{
		argv[argc++] = ptr_cmd;

		if (argc >= 10) {
			printk(KERN_ERR MODULE_NAME": too many parameters. dropping cmd.\n");
			return -EIO;
		}

		ptr_cmd = strsep(&ptr_next_tok, delimitters);
	} while (ptr_cmd != NULL);

	argc--;

	if (strcmp(argv[0], "start") == 0) {
	   if (argv[1] && strcmp(argv[1], "tcp") == 0) {
		   start_tcp_thread();
		   kspeedtest.tcp_port = simple_strtoul(argv[2], 0, 10);;
	   } else if (argv[1] && strcmp(argv[1], "udp") == 0) {
		   start_udp_thread();
		   kspeedtest.udp_port = simple_strtoul(argv[2], 0, 10);;
	   } else {
		   printk(KERN_ERR MODULE_NAME": command wrong or incomplete (usage example: 'start tcp 4711'\n");
	   }
	} else if (strcmp(argv[0], "stop") == 0) {
	   if (argv[1] && strcmp(argv[1], "tcp") == 0) {
		   stop_tcp_thread();
	   } else if (argv[1] && strcmp(argv[1], "udp") == 0) {
		   stop_udp_thread();
	   } else if (argv[1] && strcmp(argv[1], "all") == 0) {
		   stop_tcp_thread();
		   stop_udp_thread();
	   } else {
		   printk(KERN_ERR MODULE_NAME": command wrong or incomplete (usage example: 'stop tcp'\n");
	   }
	} else if (strcmp(argv[0], "verbose") == 0) {
		kspeedtest.verbose = 1;
	} else if (strcmp(argv[0], "noverbose") == 0) {
		kspeedtest.verbose = 0;
	} else {
		printk(KERN_ERR MODULE_NAME": wrong command\n");
	}
	return count;
}

static struct proc_dir_entry *dir_entry = 0;

static void __init kspeedtest_proc_init(void)
{
	struct proc_dir_entry *file_entry;

	dir_entry = proc_net_mkdir(&init_net, "kspeedtest", init_net.proc_net);
	if (dir_entry) {
		dir_entry->read_proc = 0;
		dir_entry->write_proc = 0;
	}

	file_entry = create_proc_entry("control", S_IFREG|S_IWUSR, dir_entry);
	if (file_entry) {
		file_entry->data       = NULL;
		file_entry->read_proc  = NULL;
        file_entry->write_proc = kspeedtest_write_cmds;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 32)
		file_entry->owner      = THIS_MODULE;
#endif
	}
	file_entry = proc_create("stats", S_IRUGO, dir_entry, &stats_show_fops);
	file_entry = proc_create("connections", S_IRUGO, dir_entry, 
			&connections_show_fops);
}

static void __init kspeedtest_proc_exit(void)
{
	remove_proc_entry("control", dir_entry);
	remove_proc_entry("stats", dir_entry);
	remove_proc_entry("connections", dir_entry);
	remove_proc_entry("kspeedtest", init_net.proc_net);
}

#endif

/* -------------------------------------------------------------------- */

/* init and cleanup functions */
module_init(kspeedtest_init);
module_exit(kspeedtest_exit);

/* module information */
MODULE_DESCRIPTION("kernel speedtest");

/* -------------------------------------------------------------------- */

#include <linux/module.h>
#include <linux/vermagic.h>
#include <linux/compiler.h>

MODULE_INFO(vermagic, VERMAGIC_STRING);

#undef unix
struct module __this_module
__attribute__((section(".gnu.linkonce.this_module"))) = {
// .name = __stringify(KBUILD_MODNAME),
 .name = "kspeedtest",
 .init = init_module,
#ifdef CONFIG_MODULE_UNLOAD
 .exit = cleanup_module,
#endif
};


